#ifndef ARITHMETIC_HH
#define ARITHMETIC_HH


#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/diagonalmatrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/scaledidmatrix.hh>
#include <dune/common/typetraits.hh>

/** \brief Namespace containing helper classes and functions for arithmetic operations
 *
 * Everything in this namespace is experimental and might change
 * in the near future. Especially the naming of namespace, structs,
 * and functions is not final.
 */
namespace Arithmetic
{

    /** \brief Class to identify scalar types
     *
     * Specialize this class for all types that can be used
     * like scalar quantities.
     */
    template<class T>
    struct ScalarTraits
    {
        enum { isScalar=false };
    };

    template<>
    struct ScalarTraits<float>
    {
        enum { isScalar=true };
    };

    template<>
    struct ScalarTraits<double>
    {
        enum { isScalar=true };
    };

    template<class T>
    struct ScalarTraits<Dune::FieldVector<T,1> >
    {
        enum { isScalar=true };
    };

    template<class T>
    struct ScalarTraits<Dune::FieldMatrix<T,1,1> >
    {
        enum { isScalar=true };
    };

    template<class T>
    struct ScalarTraits<Dune::DiagonalMatrix<T,1> >
    {
        enum { isScalar=true };
    };

    template<class T>
    struct ScalarTraits<Dune::ScaledIdentityMatrix<T,1> >
    {
        enum { isScalar=true };
    };


    /** \brief Class to identify matrix types and extract information
     *
     * Specialize this class for all types that can be used like a matrix.
     */
    template<class T>
    struct MatrixTraits
    {
        enum { isMatrix=false };
        enum { rows=-1};
        enum { cols=-1};
    };

    template<class T, int n, int m>
    struct MatrixTraits<Dune::FieldMatrix<T,n,m> >
    {
        enum { isMatrix=true };
        enum { rows=n};
        enum { cols=m};
    };

    template<class T, int n>
    struct MatrixTraits<Dune::DiagonalMatrix<T,n> >
    {
        enum { isMatrix=true };
        enum { rows=n};
        enum { cols=n};
    };

    template<class T, int n>
    struct MatrixTraits<Dune::ScaledIdentityMatrix<T,n> >
    {
        enum { isMatrix=true };
        enum { rows=n};
        enum { cols=n};
    };

    template<class T>
    struct MatrixTraits<Dune::BCRSMatrix<T> >
    {
        enum { isMatrix=true };
    };


    /** \brief Internal helper class for product operations
     *
     */
    template<class A, class B, class C, bool AisScalar, bool BisScalar, bool CisScalar>
    struct ProductHelper
    {
        static void addProduct(A& a, const B& b, const C& c)
        {
            b.umv(c, a);
        }
        static void subtractProduct(A& a, const B& b, const C& c)
        {
            b.mmv(c, a);
        }
    };

    // Internal helper class for scaled product operations (i.e., b is always a scalar)
    template<class A, class B, class C, class D, bool AisScalar, bool CisScalar, bool DisScalar>
    struct ScaledProductHelper
    {
        static void addProduct(A& a, const B& b, const C& c, const D& d)
        {
            c.usmv(b, d, a);
        }
        static void subtractProduct(A& a, const B& b, const C& c, const D& d)
        {
            c.usmv(-b, d, a);
        }
    };

    template<class T, int n, int m, int k>
    struct ProductHelper<Dune::FieldMatrix<T,n,k>, Dune::FieldMatrix<T,n,m>, Dune::FieldMatrix<T,m,k>, false, false, false>
    {
        typedef Dune::FieldMatrix<T,n,k> A;
        typedef Dune::FieldMatrix<T,n,m> B;
        typedef Dune::FieldMatrix<T,m,k> C;
        static void addProduct(A& a, const B& b, const C& c)
        {
            for (size_t row = 0; row < n; ++row)
            {
                for (size_t col = 0 ; col < k; ++col)
                {
                    for (size_t i = 0; i < m; ++i)
                        a[row][col] += b[row][i]*c[i][col];
                }
            }
        }
        static void subtractProduct(A& a, const B& b, const C& c)
        {
            for (size_t row = 0; row < n; ++row)
            {
                for (size_t col = 0 ; col < k; ++col)
                {
                    for (size_t i = 0; i < m; ++i)
                        a[row][col] -= b[row][i]*c[i][col];
                }
            }
        }
    };

    template<class T, int n, int m, int k>
    struct ScaledProductHelper<Dune::FieldMatrix<T,n,k>, T, Dune::FieldMatrix<T,n,m>, Dune::FieldMatrix<T,m,k>, false, false, false>
    {
        typedef Dune::FieldMatrix<T,n,k> A;
        typedef Dune::FieldMatrix<T,n,m> C;
        typedef Dune::FieldMatrix<T,m,k> D;
        static void addProduct(A& a, const T& b, const C& c, const D& d)
        {
            for (size_t row = 0; row < n; ++row) {
                for (size_t col = 0 ; col < k; ++col) {
                    for (size_t i = 0; i < m; ++i)
                        a[row][col] += b * c[row][i]*d[i][col];
                }
            }
        }
        static void subtractProduct(A& a, const T& b, const C& c, const D& d)
        {
            for (size_t row = 0; row < n; ++row) {
                for (size_t col = 0 ; col < k; ++col) {
                    for (size_t i = 0; i < m; ++i)
                        a[row][col] -= b * c[row][i]*d[i][col];
                }
            }
        }
    };

    template<class T, int n>
    struct ProductHelper<Dune::DiagonalMatrix<T,n>, Dune::DiagonalMatrix<T,n>, Dune::DiagonalMatrix<T,n>, false, false, false>
    {
        typedef Dune::DiagonalMatrix<T,n> A;
        typedef A B;
        typedef B C;
        static void addProduct(A& a, const B& b, const C& c)
        {
            for (size_t i=0; i<n; ++i)
                a.diagonal(i) += b.diagonal(i)*c.diagonal(i);
        }
        static void subtractProduct(A& a, const B& b, const C& c)
        {
            for (size_t i=0; i<n; ++i)
                a.diagonal(i) -= b.diagonal(i)*c.diagonal(i);
        }
    };

    template<class T, int n>
    struct ScaledProductHelper<Dune::DiagonalMatrix<T,n>, T, Dune::DiagonalMatrix<T,n>, Dune::DiagonalMatrix<T,n>, false, false, false>
    {
        typedef Dune::DiagonalMatrix<T,n> A;
        typedef A C;
        typedef C D;
        static void addProduct(A& a, const T& b, const C& c, const D& d)
        {
            for (size_t i=0; i<n; ++i)
                a.diagonal(i) += b * c.diagonal(i)*d.diagonal(i);
        }
        static void subtractProduct(A& a, const T& b, const C& c, const D& d)
        {
            for (size_t i=0; i<n; ++i)
                a.diagonal(i) -= b * c.diagonal(i)*d.diagonal(i);
        }
    };

    template<class T, int n>
    struct ProductHelper<Dune::FieldMatrix<T,n,n>, Dune::DiagonalMatrix<T,n>, Dune::DiagonalMatrix<T,n>, false, false, false>
    {
        typedef Dune::FieldMatrix<T,n,n> A;
        typedef Dune::DiagonalMatrix<T,n> B;
        typedef B C;
        static void addProduct(A& a, const B& b, const C& c)
        {
            for (size_t i=0; i<n; ++i)
                a[i][i] += b.diagonal(i)*c.diagonal(i);
        }
        static void subtractProduct(A& a, const B& b, const C& c)
        {
            for (size_t i=0; i<n; ++i)
                a[i][i] -= b.diagonal(i)*c.diagonal(i);
        }
    };

    template<class T, int n>
    struct ScaledProductHelper<Dune::FieldMatrix<T,n,n>, T, Dune::DiagonalMatrix<T,n>, Dune::DiagonalMatrix<T,n>, false, false, false>
    {
        typedef Dune::FieldMatrix<T,n,n> A;
        typedef Dune::DiagonalMatrix<T,n> C;
        typedef C D;
        static void addProduct(A& a, const T& b, const C& c, const D& d)
        {
            for (size_t i=0; i<n; ++i)
                a[i][i] += b * c.diagonal(i)*d.diagonal(i);
        }
        static void subtractProduct(A& a, const T& b, const C& c, const D& d)
        {
            for (size_t i=0; i<n; ++i)
                a[i][i] -= b * c.diagonal(i)*d.diagonal(i);
        }
    };

    /** \brief Specialization for b being a scalar type
      */
    template<class A, class B, class C, bool AisScalar, bool CisScalar>
    struct ProductHelper<A, B, C, AisScalar, true, CisScalar>
    {
        static void addProduct(A& a, const B& b, const C& c)
        {
            a.axpy(b, c);
        }
        static void subtractProduct(A& a, const B& b, const C& c)
        {
            a.axpy(-b, c);
        }
    };

    /** \brief Specialization for c being a scalar type
     */
    template<class A, class B, class C, class D, bool AisScalar, bool DisScalar>
    struct ScaledProductHelper<A, B, C, D, AisScalar, true, DisScalar>
    {
        static void addProduct(A& a, const B& b, const C& c, const D& d)
        {
            a.axpy(b * c, d);
        }
        static void subtractProduct(A& a, const B& b, const C& c, const D& d)
        {
            a.axpy(-b * c, d);
        }
    };

    template<class A, class B, class C>
    struct ProductHelper<A, B, C, true, true, true>
    {
        static void addProduct(A& a, const B& b, const C& c)
        {
            a += b*c;
        }
        static void subtractProduct(A& a, const B& b, const C& c)
        {
            a -= b*c;
        }
    };

    template<class A, class B, class C, class D>
    struct ScaledProductHelper<A, B, C, D, true, true, true>
    {
        static void addProduct(A& a, const B& b, const C& c, const D& d)
        {
            a += b*c*d;
        }
        static void subtractProduct(A& a, const B& b, const C& c, const D& d)
        {
            a -= b*c*d;
        }
    };

   //! Helper class for computing the cross product
    template <class T, int n>
    struct CrossProductHelper
    {
        static Dune::FieldVector<T,n> crossProduct(const Dune::FieldVector<T,n>& a, const Dune::FieldVector<T,n>& b)
        {
            DUNE_THROW(Dune::Exception, "You can only call crossProduct with dim==3");
        }
    };

    //! Specialisation for n=3
    template <class T>
    struct CrossProductHelper<T,3>
    {
        static Dune::FieldVector<T,3> crossProduct(const Dune::FieldVector<T,3>& a, const Dune::FieldVector<T,3>& b)
        {
            Dune::FieldVector<T,3> r;
            r[0] = a[1]*b[2] - a[2]*b[1];
            r[1] = a[2]*b[0] - a[0]*b[2];
            r[2] = a[0]*b[1] - a[1]*b[0];
            return r;
        }
    };

    template<class A, class TransposedA>
    struct TransposeHelper
    {
        static void transpose(const A& a, TransposedA& aT)
        {
            DUNE_THROW(Dune::Exception, "Not implemented for general matrix types!");
        }
    };

    //! Specialization for Dune::FieldMatrix
    template<class T, int n, int m>
    struct TransposeHelper<Dune::FieldMatrix<T,n,m>, Dune::FieldMatrix<T,m,n> >
    {
        static void transpose(const Dune::FieldMatrix<T,n,m>& a, Dune::FieldMatrix<T,m,n>& aT)
        {
            for (int row = 0; row < m; ++row)
                for (int col = 0 ; col < n; ++col)
                    aT[row][col] = a[col][row];
        }
    };

    //! Specialization for Dune::BCRSMatrix Type
    template<class A, class TransposedA>
    struct TransposeHelper<Dune::BCRSMatrix<A>, Dune::BCRSMatrix<TransposedA> >
    {
        static void transpose(const Dune::BCRSMatrix<A>& a, Dune::BCRSMatrix<TransposedA>& aT)
        {
            Dune::MatrixIndexSet idxSetaT(a.M(),a.N());

            typedef typename Dune::BCRSMatrix<A>::ConstColIterator ColIterator;

            // add indices into transposed matrix
            for (int row = 0; row < a.N(); ++row)
            {
                ColIterator col = a[row].begin();
                ColIterator end = a[row].end();

                for ( ; col != end; ++col)
                    idxSetaT.add(col.index(),row);
            }

            idxSetaT.exportIdx(aT);

            for (int row = 0; row < a.N(); ++row)
            {
                ColIterator col = a[row].begin();
                ColIterator end = a[row].end();

                for ( ; col != end; ++col)
                    TransposeHelper<A,TransposedA>::transpose(*col, aT[col.index()][row]);
            }
        }
    };

    /** \brief Add a product to some matrix or vector
     *
     * This function computes a+=b*c.
     *
     * This function should tolerate all meaningful
     * combinations of scalars, vectors, and matrices.
     *
     * a,b,c could be matrices with appropriate
     * dimensions. But b can also always be a scalar
     * represented by a 1-dim vector or a 1 by 1 matrix.
     */
    template<class A, class B, class C>
    void addProduct(A& a, const B& b, const C& c)
    {
        ProductHelper<A,B,C,ScalarTraits<A>::isScalar, ScalarTraits<B>::isScalar, ScalarTraits<C>::isScalar>::addProduct(a,b,c);
    }

    /** \brief Subtract a product from some matrix or vector
     *
     * This function computes a-=b*c.
     *
     * This function should tolerate all meaningful
     * combinations of scalars, vectors, and matrices.
     *
     * a,b,c could be matrices with appropriate
     * dimensions. But b can also always be a scalar
     * represented by a 1-dim vector or a 1 by 1 matrix.
     */
    template<class A, class B, class C>
    void subtractProduct(A& a, const B& b, const C& c)
    {
        ProductHelper<A,B,C,ScalarTraits<A>::isScalar, ScalarTraits<B>::isScalar, ScalarTraits<C>::isScalar>::subtractProduct(a,b,c);
    }
    //! Compute the cross product of two vectors. Only works for n==3
    template<class T, int n>
    Dune::FieldVector<T,n> crossProduct(const Dune::FieldVector<T,n>& a, const Dune::FieldVector<T,n>& b)
    {
        return CrossProductHelper<T,n>::crossProduct(a,b);
    }

    //! Compute the transposed of a matrix
    template <class A, class TransposedA>
    void transpose(const A& a, TransposedA& aT)
    {
        TransposeHelper<A,TransposedA>::transpose(a,aT);
    }

    /** \brief Add a scaled product to some matrix or vector
     *
     * This function computes a+=b*c*d.
     *
     * This function should tolerate all meaningful
     * combinations of scalars, vectors, and matrices.
     *
     * a,c,d could be matrices with appropriate dimensions. But b must
     * (currently) and c can also always be a scalar represented by a
     * 1-dim vector or a 1 by 1 matrix.
     */
    template<class A, class B, class C, class D>
    typename Dune::enable_if<ScalarTraits<B>::isScalar, void>::type
    addProduct(A& a, const B& b, const C& c, const D& d)
    {
        ScaledProductHelper<A,B,C,D,ScalarTraits<A>::isScalar, ScalarTraits<C>::isScalar, ScalarTraits<D>::isScalar>::addProduct(a,b,c,d);
    }

    /** \brief Subtract a scaled product from some matrix or vector
     *
     * This function computes a-=b*c*d.
     *
     * This function should tolerate all meaningful
     * combinations of scalars, vectors, and matrices.
     *
     * a,c,d could be matrices with appropriate dimensions. But b must
     * (currently) and c can also always be a scalar represented by a
     * 1-dim vector or a 1 by 1 matrix.
     */
    template<class A, class B, class C, class D>
    typename Dune::enable_if<ScalarTraits<B>::isScalar, void>::type
    subtractProduct(A& a, const B& b, const C& c, const D& d)
    {
        ScaledProductHelper<A,B,C,D,ScalarTraits<A>::isScalar, ScalarTraits<C>::isScalar, ScalarTraits<D>::isScalar>::subtractProduct(a,b,c,d);
    }

    /** \brief Internal helper class for Matrix operations
     *
     */
    template<class OperatorType, bool isMatrix>
    struct OperatorHelper
    {
        template <class VectorType>
        static typename VectorType::field_type
        Axy(const OperatorType &A,
            const VectorType &x,
            const VectorType &y)
        {
            VectorType tmp(y.size());
            tmp = 0.0;
            addProduct(tmp, A, x);
            return tmp * y;
        }

        template <class VectorType>
        static typename VectorType::field_type
        bmAxy(const OperatorType &A, const VectorType &b,
              const VectorType &x, const VectorType &y)
        {
            VectorType tmp = b;
            subtractProduct(tmp, A, x);
            return tmp * y;
        }
    };

    template<class MatrixType>
    struct OperatorHelper<MatrixType, true>
    {
        template <class VectorType>
        static typename VectorType::field_type
        Axy(const MatrixType &A,
            const VectorType &x,
            const VectorType &y)
        {
            assert(x.N() == A.M());
            assert(y.N() == A.N());

            typename VectorType::field_type outer = 0.0;
            typename VectorType::block_type inner;
            typename MatrixType::ConstIterator endi=A.end();
            for (typename MatrixType::ConstIterator i=A.begin(); i!=endi; ++i) {
                inner = 0.0;
                const size_t iindex = i.index();
                typename MatrixType::row_type::ConstIterator endj = i->end();
                for (typename MatrixType::row_type::ConstIterator j=i->begin();
                     j!=endj; ++j)
                    addProduct(inner, *j, x[j.index()]);
                outer += inner * y[iindex];
            }
            return outer;
        }

        template <class VectorType>
        static typename VectorType::field_type
        bmAxy(const MatrixType &A, const VectorType &b,
              const VectorType &x, const VectorType &y)
        {
            assert(x.N() == A.M());
            assert(y.N() == A.N());
            assert(y.N() == b.N());

            typename VectorType::field_type outer = 0.0;
            typename VectorType::block_type inner;
            typename MatrixType::ConstIterator endi=A.end();
            for (typename MatrixType::ConstIterator i=A.begin(); i!=endi; ++i) {
                const size_t iindex = i.index();
                inner = b[iindex];
                typename MatrixType::row_type::ConstIterator endj = i->end();
                for (typename MatrixType::row_type::ConstIterator j=i->begin();
                     j!=endj; ++j)
                    subtractProduct(inner, *j, x[j.index()]);
                outer += inner * y[iindex];
            }
            return outer;
        }
    };

    //! Compute \f$(Ax,y)\f$
    template <class OperatorType, class VectorType>
    typename VectorType::field_type
    Axy(const OperatorType &A,
        const VectorType &x,
        const VectorType &y)
    {
        return OperatorHelper<OperatorType,
                              MatrixTraits<OperatorType>::isMatrix>::Axy(A, x, y);
    }

    //! Compute \f$(b-Ax,y)\f$
    template <class OperatorType, class VectorType>
    typename VectorType::field_type
    bmAxy(const OperatorType &A,
          const VectorType &b,
          const VectorType &x,
          const VectorType &y)
    {
        return OperatorHelper<OperatorType,
                              MatrixTraits<OperatorType>::isMatrix>::bmAxy(A, b, x, y);
    }
};

#endif
